import pytest
from tests.utils import *
from tests.constants import *

ADD_TIMEOUT = 600
uid = "test_insert"
field_name = default_float_vec_field_name
binary_field_name = default_binary_vec_field_name
default_single_query = {
    "bool": {
        "must": [
            {"vector": {field_name: {"topk": 10, "query": gen_vectors(1, default_dim), "metric_type": "L2",
                                     "params": {"nprobe": 10}}}}
        ]
    }
}


class TestInsertBase:
    """
    ******************************************************************
      The following cases are used to test `insert` function
    ******************************************************************
    """

    @pytest.fixture(
        scope="function",
        params=gen_simple_index()
    )
    def get_simple_index(self, request, connect):
        if str(connect._cmd("mode")) == "CPU":
            if request.param["index_type"] in index_cpu_not_support():
                pytest.skip("CPU not support index_type: ivf_sq8h")
        return request.param

    @pytest.fixture(
        scope="function",
        params=gen_single_filter_fields()
    )
    def get_filter_field(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_single_vector_fields()
    )
    def get_vector_field(self, request):
        yield request.param

    def test_add_vector_with_empty_vector(self, connect, collection):
        '''
        target: test add vectors with empty vectors list
        method: set empty vectors list as add method params
        expected: raises a Exception
        '''
        vector = []
        with pytest.raises(Exception) as e:
            status, ids = connect.bulk_insert(collection, vector)

    def test_add_vector_with_None(self, connect, collection):
        '''
        target: test add vectors with None
        method: set None as add method params
        expected: raises a Exception
        '''
        vector = None
        with pytest.raises(Exception) as e:
            status, ids = connect.bulk_insert(collection, vector)

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_collection_not_existed(self, connect):
        '''
        target: test insert, with collection not existed
        method: insert entity into a random named collection
        expected: error raised 
        '''
        collection_name = gen_unique_str(uid)
        with pytest.raises(Exception) as e:
            connect.bulk_insert(collection_name, default_entities)

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_drop_collection(self, connect, collection):
        '''
        target: test delete collection after insert vector
        method: insert vector and delete collection
        expected: no error raised
        '''
        ids = connect.bulk_insert(collection, default_entity)
        assert len(ids) == 1
        connect.drop_collection(collection)

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_sleep_drop_collection(self, connect, collection):
        '''
        target: test delete collection after insert vector for a while
        method: insert vector, sleep, and delete collection
        expected: no error raised 
        '''
        ids = connect.bulk_insert(collection, default_entity)
        assert len(ids) == 1
        connect.flush([collection])
        connect.drop_collection(collection)

    @pytest.mark.skip("create index")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_create_index(self, connect, collection, get_simple_index):
        '''
        target: test build index insert after vector
        method: insert vector and build index
        expected: no error raised
        '''
        ids = connect.bulk_insert(collection, default_entities)
        assert len(ids) == default_nb
        connect.flush([collection])
        connect.create_index(collection, field_name, get_simple_index)
        info = connect.describe_collection(collection)
        fields = info["fields"]
        for field in fields:
            if field["name"] == field_name:
                assert field["indexes"][0] == get_simple_index

    @pytest.mark.skip("create index")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_create_index_new(self, connect, collection, get_simple_index):
        '''
        target: test build index insert after vector
        method: insert vector and build index
        expected: no error raised
        '''
        ids = connect.bulk_insert(collection, default_entities_new)
        assert len(ids) == default_nb
        connect.flush([collection])
        connect.create_index(collection, field_name, get_simple_index)
        info = connect.describe_collection(collection)
        fields = info["fields"]
        for field in fields:
            if field["name"] == field_name:
                assert field["indexes"][0] == get_simple_index

    @pytest.mark.skip("create index")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_after_create_index(self, connect, collection, get_simple_index):
        '''
        target: test build index insert after vector
        method: insert vector and build index
        expected: no error raised
        '''
        connect.create_index(collection, field_name, get_simple_index)
        ids = connect.bulk_insert(collection, default_entities)
        assert len(ids) == default_nb
        info = connect.describe_collection(collection)
        fields = info["fields"]
        for field in fields:
            if field["name"] == field_name:
                assert field["indexes"][0] == get_simple_index

    # @pytest.mark.skip(" later ")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_search(self, connect, collection):
        '''
        target: test search vector after insert vector after a while
        method: insert vector, sleep, and search collection
        expected: no error raised 
        '''
        ids = connect.bulk_insert(collection, default_entities)
        connect.flush([collection])
        res = connect.search(collection, default_single_query)
        logging.getLogger().debug(res)
        assert res
    
    @pytest.mark.skip("segment row count")
    def test_insert_segment_row_count(self, connect, collection):
        nb = default_segment_row_limit + 1
        res_ids = connect.bulk_insert(collection, gen_entities(nb))
        connect.flush([collection])
        assert len(res_ids) == nb
        stats = connect.get_collection_stats(collection)
        assert len(stats['partitions'][0]['segments']) == 2
        for segment in stats['partitions'][0]['segments']:
            assert segment['row_count'] in [default_segment_row_limit, 1]

    @pytest.fixture(
        scope="function",
        params=[
            1,
            2000
        ],
    )
    def insert_count(self, request):
        yield request.param

    @pytest.mark.skip(" todo support count entities")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_ids(self, connect, id_collection, insert_count):
        '''
        target: test insert vectors in collection, use customize ids
        method: create collection and insert vectors in it, check the ids returned and the collection length after vectors inserted
        expected: the length of ids and the collection row count
        '''
        nb = insert_count
        ids = [i for i in range(nb)]
        res_ids = connect.bulk_insert(id_collection, gen_entities(nb), ids)
        connect.flush([id_collection])
        assert len(res_ids) == nb
        assert res_ids == ids
        res_count = connect.count_entities(id_collection)
        assert res_count == nb

    @pytest.mark.skip(" todo support count entities")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_the_same_ids(self, connect, id_collection, insert_count):
        '''
        target: test insert vectors in collection, use customize the same ids
        method: create collection and insert vectors in it, check the ids returned and the collection length after vectors inserted
        expected: the length of ids and the collection row count
        '''
        nb = insert_count
        ids = [1 for i in range(nb)]
        res_ids = connect.bulk_insert(id_collection, gen_entities(nb), ids)
        connect.flush([id_collection])
        assert len(res_ids) == nb
        assert res_ids == ids
        res_count = connect.count_entities(id_collection)
        assert res_count == nb

    @pytest.mark.skip(" todo support count entities")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_ids_fields(self, connect, get_filter_field, get_vector_field):
        '''
        target: test create normal collection with different fields, insert entities into id with ids
        method: create collection with diff fields: metric/field_type/..., insert, and count
        expected: row count correct
        '''
        nb = 5
        filter_field = get_filter_field
        vector_field = get_vector_field
        collection_name = gen_unique_str("test_collection")
        fields = {
            "fields": [filter_field, vector_field],
            "segment_row_limit": default_segment_row_limit,
            "auto_id": True
        }
        connect.create_collection(collection_name, fields)
        ids = [i for i in range(nb)]
        entities = gen_entities_by_fields(fields["fields"], nb, default_dim)
        res_ids = connect.bulk_insert(collection_name, entities, ids)
        assert res_ids == ids
        connect.flush([collection_name])
        res_count = connect.count_entities(collection_name)
        assert res_count == nb

    # TODO: assert exception && enable
    @pytest.mark.level(2)
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_twice_ids_no_ids(self, connect, id_collection):
        '''
        target: check the result of insert, with params ids and no ids
        method: test insert vectors twice, use customize ids first, and then use no ids
        expected:  error raised
        '''
        ids = [i for i in range(default_nb)]
        res_ids = connect.bulk_insert(id_collection, default_entities, ids)
        with pytest.raises(Exception) as e:
            res_ids_new = connect.bulk_insert(id_collection, default_entities)

    # TODO: assert exception && enable
    @pytest.mark.level(2)
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_twice_not_ids_ids(self, connect, id_collection):
        '''
        target: check the result of insert, with params ids and no ids
        method: test insert vectors twice, use not ids first, and then use customize ids
        expected:  error raised
        '''
        with pytest.raises(Exception) as e:
            res_ids = connect.bulk_insert(id_collection, default_entities)

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_ids_length_not_match_batch(self, connect, id_collection):
        '''
        target: test insert vectors in collection, use customize ids, len(ids) != len(vectors)
        method: create collection and insert vectors in it
        expected: raise an exception
        '''
        ids = [i for i in range(1, default_nb)]
        logging.getLogger().info(len(ids))
        with pytest.raises(Exception) as e:
            res_ids = connect.bulk_insert(id_collection, default_entities, ids)

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_ids_length_not_match_single(self, connect, collection):
        '''
        target: test insert vectors in collection, use customize ids, len(ids) != len(vectors)
        method: create collection and insert vectors in it
        expected: raise an exception
        '''
        ids = [i for i in range(1, default_nb)]
        logging.getLogger().info(len(ids))
        with pytest.raises(Exception) as e:
            res_ids = connect.bulk_insert(collection, default_entity, ids)

    @pytest.mark.skip(" todo support count entities")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_ids_fields(self, connect, get_filter_field, get_vector_field):
        '''
        target: test create normal collection with different fields, insert entities into id without ids
        method: create collection with diff fields: metric/field_type/..., insert, and count
        expected: row count correct
        '''
        nb = 5
        filter_field = get_filter_field
        vector_field = get_vector_field
        collection_name = gen_unique_str("test_collection")
        fields = {
            "fields": [filter_field, vector_field],
            "segment_row_limit": default_segment_row_limit
        }
        connect.create_collection(collection_name, fields)
        entities = gen_entities_by_fields(fields["fields"], nb, default_dim)
        res_ids = connect.bulk_insert(collection_name, entities)
        connect.flush([collection_name])
        res_count = connect.count_entities(collection_name)
        assert res_count == nb

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_tag(self, connect, collection):
        '''
        target: test insert entities in collection created before
        method: create collection and insert entities in it, with the partition_tag param
        expected: the collection row count equals to nq
        '''
        connect.create_partition(collection, default_tag)
        ids = connect.bulk_insert(collection, default_entities, partition_tag=default_tag)
        assert len(ids) == default_nb
        assert connect.has_partition(collection, default_tag)

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_tag_with_ids(self, connect, id_collection):
        '''
        target: test insert entities in collection created before, insert with ids
        method: create collection and insert entities in it, with the partition_tag param
        expected: the collection row count equals to nq
        '''
        connect.create_partition(id_collection, default_tag)
        ids = [i for i in range(default_nb)]
        res_ids = connect.bulk_insert(id_collection, default_entities, ids, partition_tag=default_tag)
        assert res_ids == ids


    @pytest.mark.skip(" todo support count entities")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_default_tag(self, connect, collection):
        '''
        target: test insert entities into default partition
        method: create partition and insert info collection without tag params
        expected: the collection row count equals to nb
        '''
        connect.create_partition(collection, default_tag)
        ids = connect.bulk_insert(collection, default_entities)
        connect.flush([collection])
        assert len(ids) == default_nb
        res_count = connect.count_entities(collection)
        assert res_count == default_nb

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_tag_not_existed(self, connect, collection):
        '''
        target: test insert entities in collection created before
        method: create collection and insert entities in it, with the not existed partition_tag param
        expected: error raised
        '''
        tag = gen_unique_str()
        with pytest.raises(Exception) as e:
            ids = connect.bulk_insert(collection, default_entities, partition_tag=tag)

    @pytest.mark.skip(" not support count entities")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_tag_existed(self, connect, collection):
        '''
        target: test insert entities in collection created before
        method: create collection and insert entities in it repeatly, with the partition_tag param
        expected: the collection row count equals to nq
        '''
        connect.create_partition(collection, default_tag)
        ids = connect.bulk_insert(collection, default_entities, partition_tag=default_tag)
        ids = connect.bulk_insert(collection, default_entities, partition_tag=default_tag)
        connect.flush([collection])
        res_count = connect.count_entities(collection)
        assert res_count == 2 * default_nb

    @pytest.mark.level(2)
    def test_insert_without_connect(self, dis_connect, collection):
        '''
        target: test insert entities without connection
        method: create collection and insert entities in it, check if inserted successfully
        expected: raise exception
        '''
        with pytest.raises(Exception) as e:
            ids = dis_connect.bulk_insert(collection, default_entities)

    def test_insert_collection_not_existed(self, connect):
        '''
        target: test insert entities in collection, which not existed before
        method: insert entities collection not existed, check the status
        expected: error raised
        '''
        with pytest.raises(Exception) as e:
            ids = connect.bulk_insert(gen_unique_str("not_exist_collection"), default_entities)

    @pytest.mark.skip("to do add dim check ")
    def test_insert_dim_not_matched(self, connect, collection):
        '''
        target: test insert entities, the vector dimension is not equal to the collection dimension
        method: the entities dimension is half of the collection dimension, check the status
        expected: error raised
        '''
        vectors = gen_vectors(default_nb, int(default_dim) // 2)
        insert_entities = copy.deepcopy(default_entities)
        insert_entities[-1]["values"] = vectors
        with pytest.raises(Exception) as e:
            ids = connect.bulk_insert(collection, insert_entities)


    def test_insert_with_field_name_not_match(self, connect, collection):
        '''
        target: test insert entities, with the entity field name updated
        method: update entity field name
        expected: error raised
        '''
        tmp_entity = update_field_name(copy.deepcopy(default_entity), "int64", "int64new")
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    @pytest.mark.skip(" todo support  type check")
    def test_insert_with_field_type_not_match(self, connect, collection):
        '''
        target: test insert entities, with the entity field type updated
        method: update entity field type
        expected: error raised
        '''
        tmp_entity = update_field_type(copy.deepcopy(default_entity), "int64", DataType.FLOAT)
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    @pytest.mark.skip("to do add field_type check ")
    @pytest.mark.level(2)
    def test_insert_with_field_type_not_match_B(self, connect, collection):
        '''
        target: test insert entities, with the entity field type updated
        method: update entity field type
        expected: error raised
        '''
        tmp_entity = update_field_type(copy.deepcopy(default_entity), "int64", DataType.DOUBLE)
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    @pytest.mark.level(2)
    def test_insert_with_field_value_not_match(self, connect, collection):
        '''
        target: test insert entities, with the entity field value updated
        method: update entity field value
        expected: error raised
        '''
        tmp_entity = update_field_value(copy.deepcopy(default_entity), DataType.FLOAT, 's')
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    def test_insert_with_field_more(self, connect, collection):
        '''
        target: test insert entities, with more fields than collection schema
        method: add entity field
        expected: error raised
        '''
        tmp_entity = add_field(copy.deepcopy(default_entity))
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    def test_insert_with_field_vector_more(self, connect, collection):
        '''
        target: test insert entities, with more fields than collection schema
        method: add entity vector field
        expected: error raised
        '''
        tmp_entity = add_vector_field(default_nb, default_dim)
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    def test_insert_with_field_less(self, connect, collection):
        '''
        target: test insert entities, with less fields than collection schema
        method: remove entity field
        expected: error raised
        '''
        tmp_entity = remove_field(copy.deepcopy(default_entity))
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    def test_insert_with_field_vector_less(self, connect, collection):
        '''
        target: test insert entities, with less fields than collection schema
        method: remove entity vector field
        expected: error raised
        '''
        tmp_entity = remove_vector_field(copy.deepcopy(default_entity))
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    def test_insert_with_no_field_vector_value(self, connect, collection):
        '''
        target: test insert entities, with no vector field value
        method: remove entity vector field
        expected: error raised
        '''
        tmp_entity = copy.deepcopy(default_entity)
        del tmp_entity[-1]["values"]
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    def test_insert_with_no_field_vector_type(self, connect, collection):
        '''
        target: test insert entities, with no vector field type
        method: remove entity vector field
        expected: error raised
        '''
        tmp_entity = copy.deepcopy(default_entity)
        del tmp_entity[-1]["type"]
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    def test_insert_with_no_field_vector_name(self, connect, collection):
        '''
        target: test insert entities, with no vector field name
        method: remove entity vector field
        expected: error raised
        '''
        tmp_entity = copy.deepcopy(default_entity)
        del tmp_entity[-1]["name"]
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    @pytest.mark.skip("support count entities")
    @pytest.mark.level(2)
    @pytest.mark.timeout(30)
    def test_collection_insert_rows_count_multi_threading(self, args, collection):
        '''
        target: test collection rows_count is correct or not with multi threading
        method: create collection and insert entities in it(idmap),
            assert the value returned by count_entities method is equal to length of entities
        expected: the count is equal to the length of entities
        '''
        if args["handler"] == "HTTP":
            pytest.skip("Skip test in http mode")
        thread_num = 8
        threads = []
        milvus = get_milvus(host=args["ip"], port=args["port"], handler=args["handler"], try_connect=False)

        def insert(thread_i):
            logging.getLogger().info("In thread-%d" % thread_i)
            milvus.bulk_insert(collection, default_entities)
            milvus.flush([collection])

        for i in range(thread_num):
            t = MilvusTestThread(target=insert, args=(i,))
            threads.append(t)
            t.start()
        for t in threads:
            t.join()
        res_count = milvus.count_entities(collection)
        assert res_count == thread_num * default_nb

    # TODO: unable to set config
    @pytest.mark.skip("get entity by id")
    @pytest.mark.level(2)
    def _test_insert_disable_auto_flush(self, connect, collection):
        '''
        target: test insert entities, with disable autoflush
        method: disable autoflush and insert, get entity
        expected: the count is equal to 0
        '''
        delete_nums = 500
        disable_flush(connect)
        ids = connect.bulk_insert(collection, default_entities)
        res = connect.get_entity_by_id(collection, ids[:delete_nums])
        assert len(res) == delete_nums
        assert res[0] is None



class TestInsertBinary:
    @pytest.fixture(
        scope="function",
        params=gen_binary_index()
    )
    def get_binary_index(self, request):
        request.param["metric_type"] = "JACCARD"
        return request.param

    @pytest.mark.skip("count entities")
    def test_insert_binary_entities(self, connect, binary_collection):
        '''
        target: test insert entities in binary collection
        method: create collection and insert binary entities in it
        expected: the collection row count equals to nb
        '''
        ids = connect.bulk_insert(binary_collection, default_binary_entities)
        assert len(ids) == default_nb
        connect.flush()
        assert connect.count_entities(binary_collection) == default_nb

    @pytest.mark.skip("count entities")
    def test_insert_binary_entities_new(self, connect, binary_collection):
        '''
        target: test insert entities in binary collection
        method: create collection and insert binary entities in it
        expected: the collection row count equals to nb
        '''
        ids = connect.bulk_insert(binary_collection, default_binary_entities_new)
        assert len(ids) == default_nb
        connect.flush()
        assert connect.count_entities(binary_collection) == default_nb

    # @pytest.mark.skip
    def test_insert_binary_tag(self, connect, binary_collection):
        '''
        target: test insert entities and create partition tag
        method: create collection and insert binary entities in it, with the partition_tag param
        expected: the collection row count equals to nb
        '''
        connect.create_partition(binary_collection, default_tag)
        ids = connect.bulk_insert(binary_collection, default_binary_entities, partition_tag=default_tag)
        assert len(ids) == default_nb
        assert connect.has_partition(binary_collection, default_tag)

    @pytest.mark.skip("count entities")
    @pytest.mark.level(2)
    def test_insert_binary_multi_times(self, connect, binary_collection):
        '''
        target: test insert entities multi times and final flush
        method: create collection and insert binary entity multi and final flush
        expected: the collection row count equals to nb
        '''
        for i in range(default_nb):
            ids = connect.bulk_insert(binary_collection, default_binary_entity)
            assert len(ids) == 1
        connect.flush([binary_collection])
        assert connect.count_entities(binary_collection) == default_nb

    @pytest.mark.skip("create index")
    def test_insert_binary_after_create_index(self, connect, binary_collection, get_binary_index):
        '''
        target: test insert binary entities after build index
        method: build index and insert entities
        expected: no error raised
        '''
        connect.create_index(binary_collection, binary_field_name, get_binary_index)
        ids = connect.bulk_insert(binary_collection, default_binary_entities)
        assert len(ids) == default_nb
        connect.flush([binary_collection])
        info = connect.describe_collection(binary_collection)
        fields = info["fields"]
        for field in fields:
            if field["name"] == binary_field_name:
                assert field["indexes"][0] == get_binary_index

    @pytest.mark.skip("create index")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_binary_create_index(self, connect, binary_collection, get_binary_index):
        '''
        target: test build index insert after vector
        method: insert vector and build index
        expected: no error raised
        '''
        ids = connect.bulk_insert(binary_collection, default_binary_entities)
        assert len(ids) == default_nb
        connect.flush([binary_collection])
        connect.create_index(binary_collection, binary_field_name, get_binary_index)
        info = connect.describe_collection(binary_collection)
        fields = info["fields"]
        for field in fields:
            if field["name"] == binary_field_name:
                assert field["indexes"][0] == get_binary_index

    @pytest.mark.skip("binary search")
    def test_insert_binary_search(self, connect, binary_collection):
        '''
        target: test search vector after insert vector after a while
        method: insert vector, sleep, and search collection
        expected: no error raised
        '''
        ids = connect.bulk_insert(binary_collection, default_binary_entities)
        connect.flush([binary_collection])
        query, vecs = gen_query_vectors(binary_field_name, default_binary_entities, default_top_k, 1, metric_type="JACCARD")
        res = connect.search(binary_collection, query)
        logging.getLogger().debug(res)
        assert res


class TestInsertAsync:
    @pytest.fixture(scope="function", autouse=True)
    def skip_http_check(self, args):
        if args["handler"] == "HTTP":
            pytest.skip("skip in http mode")

    @pytest.fixture(
        scope="function",
        params=[
            1,
            1000
        ],
    )
    def insert_count(self, request):
        yield request.param

    def check_status(self, result):
        logging.getLogger().info("In callback check status")
        assert not result

    def check_result(self, result):
        logging.getLogger().info("In callback check status")
        assert result

    def test_insert_async(self, connect, collection, insert_count):
        '''
        target: test insert vectors with different length of vectors
        method: set different vectors as insert method params
        expected: length of ids is equal to the length of vectors
        '''
        nb = insert_count
        future = connect.bulk_insert(collection, gen_entities(nb), _async=True)
        ids = future.result()
        connect.flush([collection])
        assert len(ids) == nb

    @pytest.mark.level(2)
    def test_insert_async_false(self, connect, collection, insert_count):
        '''
        target: test insert vectors with different length of vectors
        method: set different vectors as insert method params
        expected: length of ids is equal to the length of vectors
        '''
        nb = insert_count
        ids = connect.bulk_insert(collection, gen_entities(nb), _async=False)
        # ids = future.result()
        connect.flush([collection])
        assert len(ids) == nb

    def test_insert_async_callback(self, connect, collection, insert_count):
        '''
        target: test insert vectors with different length of vectors
        method: set different vectors as insert method params
        expected: length of ids is equal to the length of vectors
        '''
        nb = insert_count
        future = connect.bulk_insert(collection, gen_entities(nb), _async=True, _callback=self.check_result)
        future.done()
        ids = future.result()
        assert len(ids) == nb


    @pytest.mark.skip("count entities")
    @pytest.mark.level(2)
    def test_insert_async_long(self, connect, collection):
        '''
        target: test insert vectors with different length of vectors
        method: set different vectors as insert method params
        expected: length of ids is equal to the length of vectors
        '''
        nb = 50000
        future = connect.bulk_insert(collection, gen_entities(nb), _async=True, _callback=self.check_result)
        result = future.result()
        assert len(result) == nb
        connect.flush([collection])
        count = connect.count_entities(collection)
        logging.getLogger().info(count)
        assert count == nb

    @pytest.mark.skip("count entities")
    @pytest.mark.level(2)
    def test_insert_async_callback_timeout(self, connect, collection):
        '''
        target: test insert vectors with different length of vectors
        method: set different vectors as insert method params
        expected: length of ids is equal to the length of vectors
        '''
        nb = 100000
        future = connect.bulk_insert(collection, gen_entities(nb), _async=True, _callback=self.check_status, timeout=1)
        with pytest.raises(Exception) as e:
            result = future.result()
        count = connect.count_entities(collection)
        assert count == 0

    def test_insert_async_invalid_params(self, connect):
        '''
        target: test insert vectors with different length of vectors
        method: set different vectors as insert method params
        expected: raise exception
        '''
        collection_new = gen_unique_str()
        with pytest.raises(Exception) as e:
            future = connect.bulk_insert(collection_new, default_entities, _async=True)
            result = future.result()

    def test_insert_async_invalid_params_raise_exception(self, connect, collection):
        '''
        target: test insert vectors with different length of vectors
        method: set different vectors as insert method params
        expected: raise exception
        '''
        entities = []
        with pytest.raises(Exception) as e:
            future = connect.bulk_insert(collection, entities, _async=True)
            future.result()


class TestInsertMultiCollections:
    """
    ******************************************************************
      The following cases are used to test `insert` function
    ******************************************************************
    """

    @pytest.fixture(
        scope="function",
        params=gen_simple_index()
    )
    def get_simple_index(self, request, connect):
        logging.getLogger().info(request.param)
        if str(connect._cmd("mode")) == "CPU":
            if request.param["index_type"] in index_cpu_not_support():
                pytest.skip("sq8h not support in CPU mode")
        return request.param

    @pytest.mark.skip("count entities")
    def test_insert_vector_multi_collections(self, connect):
        '''
        target: test insert entities
        method: create 10 collections and insert entities into them in turn
        expected: row count
        '''
        collection_num = 10
        collection_list = []
        for i in range(collection_num):
            collection_name = gen_unique_str(uid)
            collection_list.append(collection_name)
            connect.create_collection(collection_name, default_fields)
            ids = connect.bulk_insert(collection_name, default_entities)
            connect.flush([collection_name])
            assert len(ids) == default_nb
            count = connect.count_entities(collection_name)
            assert count == default_nb

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_drop_collection_insert_vector_another(self, connect, collection):
        '''
        target: test insert vector to collection_1 after collection_2 deleted
        method: delete collection_2 and insert vector to collection_1
        expected: row count equals the length of entities inserted
        '''
        collection_name = gen_unique_str(uid)
        connect.create_collection(collection_name, default_fields)
        connect.drop_collection(collection)
        ids = connect.bulk_insert(collection_name, default_entity)
        connect.flush([collection_name])
        assert len(ids) == 1

    @pytest.mark.skip("create index")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_create_index_insert_vector_another(self, connect, collection, get_simple_index):
        '''
        target: test insert vector to collection_2 after build index for collection_1
        method: build index and insert vector
        expected: status ok
        '''
        collection_name = gen_unique_str(uid)
        connect.create_collection(collection_name, default_fields)
        connect.create_index(collection, field_name, get_simple_index)
        ids = connect.bulk_insert(collection, default_entity)
        connect.drop_collection(collection_name)

    @pytest.mark.skip("create index")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_vector_create_index_another(self, connect, collection, get_simple_index):
        '''
        target: test insert vector to collection_2 after build index for collection_1
        method: build index and insert vector
        expected: status ok
        '''
        collection_name = gen_unique_str(uid)
        connect.create_collection(collection_name, default_fields)
        ids = connect.bulk_insert(collection, default_entity)
        connect.create_index(collection, field_name, get_simple_index)
        count = connect.count_entities(collection_name)
        assert count == 0

    @pytest.mark.skip("create index")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_vector_sleep_create_index_another(self, connect, collection, get_simple_index):
        '''
        target: test insert vector to collection_2 after build index for collection_1 for a while
        method: build index and insert vector
        expected: status ok
        '''
        collection_name = gen_unique_str(uid)
        connect.create_collection(collection_name, default_fields)
        ids = connect.bulk_insert(collection, default_entity)
        connect.flush([collection])
        connect.create_index(collection, field_name, get_simple_index)
        count = connect.count_entities(collection)
        assert count == 1

    @pytest.mark.skip("count entities")
    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_search_vector_insert_vector_another(self, connect, collection):
        '''
        target: test insert vector to collection_1 after search collection_2
        method: search collection and insert vector
        expected: status ok
        '''
        collection_name = gen_unique_str(uid)
        connect.create_collection(collection_name, default_fields)
        res = connect.search(collection, default_single_query)
        logging.getLogger().debug(res)
        ids = connect.bulk_insert(collection_name, default_entity)
        connect.flush()
        count = connect.count_entities(collection_name)
        assert count == 1

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_vector_search_vector_another(self, connect, collection):
        '''
        target: test insert vector to collection_1 after search collection_2
        method: search collection and insert vector
        expected: status ok
        '''
        collection_name = gen_unique_str(uid)
        connect.create_collection(collection_name, default_fields)
        ids = connect.bulk_insert(collection, default_entity)
        result = connect.search(collection_name, default_single_query)

    @pytest.mark.timeout(ADD_TIMEOUT)
    def test_insert_vector_sleep_search_vector_another(self, connect, collection):
        '''
        target: test insert vector to collection_1 after search collection_2 a while
        method: search collection , sleep, and insert vector
        expected: status ok
        '''
        collection_name = gen_unique_str(uid)
        connect.create_collection(collection_name, default_fields)
        ids = connect.bulk_insert(collection, default_entity)
        connect.flush([collection])
        result = connect.search(collection_name, default_single_query)


class TestInsertInvalid(object):
    """
    Test inserting vectors with invalid collection names
    """

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_collection_name(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_tag_name(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_field_name(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_field_type(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_field_int_value(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_ints()
    )
    def get_entity_id(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_vectors()
    )
    def get_field_vectors_value(self, request):
        yield request.param

    def test_insert_ids_invalid(self, connect, id_collection, get_entity_id):
        '''
        target: test insert, with using customize ids, which are not int64
        method: create collection and insert entities in it
        expected: raise an exception
        '''
        entity_id = get_entity_id
        ids = [entity_id for _ in range(default_nb)]
        with pytest.raises(Exception):
            connect.bulk_insert(id_collection, default_entities, ids)

    def test_insert_with_invalid_collection_name(self, connect, get_collection_name):
        collection_name = get_collection_name
        with pytest.raises(Exception):
            connect.bulk_insert(collection_name, default_entity)

    def test_insert_with_invalid_tag_name(self, connect, collection, get_tag_name):
        tag_name = get_tag_name
        connect.create_partition(collection, default_tag)
        if tag_name is not None:
            with pytest.raises(Exception):
                connect.bulk_insert(collection, default_entity, partition_tag=tag_name)
        else:
            connect.bulk_insert(collection, default_entity, partition_tag=tag_name)

    def test_insert_with_invalid_field_name(self, connect, collection, get_field_name):
        field_name = get_field_name
        tmp_entity = update_field_name(copy.deepcopy(default_entity), "int64", get_field_name)
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    @pytest.mark.skip("laster add check of field type")
    def test_insert_with_invalid_field_type(self, connect, collection, get_field_type):
        field_type = get_field_type
        tmp_entity = update_field_type(copy.deepcopy(default_entity), 'float', field_type)
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    @pytest.mark.skip("laster add check of field value")
    def test_insert_with_invalid_field_value(self, connect, collection, get_field_int_value):
        field_value = get_field_int_value
        tmp_entity = update_field_type(copy.deepcopy(default_entity), 'int64', field_value)
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)

    def test_insert_with_invalid_field_vector_value(self, connect, collection, get_field_vectors_value):
        tmp_entity = copy.deepcopy(default_entity)
        src_vector = tmp_entity[-1]["values"]
        src_vector[0][1] = get_field_vectors_value
        with pytest.raises(Exception):
            connect.bulk_insert(collection, tmp_entity)


class TestInsertInvalidBinary(object):
    """
    Test inserting vectors with invalid collection names
    """

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_collection_name(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_tag_name(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_field_name(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_field_type(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_strs()
    )
    def get_field_int_value(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_ints()
    )
    def get_entity_id(self, request):
        yield request.param

    @pytest.fixture(
        scope="function",
        params=gen_invalid_vectors()
    )
    def get_field_vectors_value(self, request):
        yield request.param

    @pytest.mark.level(2)
    def test_insert_with_invalid_field_name(self, connect, binary_collection, get_field_name):
        tmp_entity = update_field_name(copy.deepcopy(default_binary_entity), "int64", get_field_name)
        with pytest.raises(Exception):
            connect.bulk_insert(binary_collection, tmp_entity)

    @pytest.mark.skip("todo support row data check")
    @pytest.mark.level(2)
    def test_insert_with_invalid_field_value(self, connect, binary_collection, get_field_int_value):
        tmp_entity = update_field_type(copy.deepcopy(default_binary_entity), 'int64', get_field_int_value)
        with pytest.raises(Exception):
            connect.bulk_insert(binary_collection, tmp_entity)

    @pytest.mark.skip("todo support row data check")
    @pytest.mark.level(2)
    def test_insert_with_invalid_field_vector_value(self, connect, binary_collection, get_field_vectors_value):
        tmp_entity = copy.deepcopy(default_binary_entity)
        src_vector = tmp_entity[-1]["values"]
        src_vector[0][1] = get_field_vectors_value
        with pytest.raises(Exception):
            connect.bulk_insert(binary_collection, tmp_entity)

    @pytest.mark.level(2)
    def test_insert_ids_invalid(self, connect, binary_id_collection, get_entity_id):
        '''
        target: test insert, with using customize ids, which are not int64
        method: create collection and insert entities in it
        expected: raise an exception
        '''
        entity_id = get_entity_id
        ids = [entity_id for _ in range(default_nb)]
        with pytest.raises(Exception):
            connect.bulk_insert(binary_id_collection, default_binary_entities, ids)

    @pytest.mark.skip("check filed")
    @pytest.mark.level(2)
    def test_insert_with_invalid_field_type(self, connect, binary_collection, get_field_type):
        field_type = get_field_type
        tmp_entity = update_field_type(copy.deepcopy(default_binary_entity), 'int64', field_type)
        with pytest.raises(Exception):
            connect.bulk_insert(binary_collection, tmp_entity)

    @pytest.mark.skip("check field")
    @pytest.mark.level(2)
    def test_insert_with_invalid_field_vector_value(self, connect, binary_collection, get_field_vectors_value):
        tmp_entity = copy.deepcopy(default_binary_entities)
        src_vector = tmp_entity[-1]["values"]
        src_vector[1] = get_field_vectors_value
        with pytest.raises(Exception):
            connect.bulk_insert(binary_collection, tmp_entity)



from tests import utils

default_fields = utils.gen_default_fields()
default_binary_fields = utils.gen_binary_default_fields()

default_entity = utils.gen_entities(1)
default_raw_binary_vector, default_binary_entity = utils.gen_binary_entities(1)

default_entity_row = utils.gen_entities_rows(1)
default_raw_binary_vector_row, default_binary_entity_row = utils.gen_binary_entities_rows(1)


default_entities = utils.gen_entities(utils.default_nb)
default_raw_binary_vectors, default_binary_entities = utils.gen_binary_entities(utils.default_nb)


default_entities_new = utils.gen_entities_new(utils.default_nb)
default_raw_binary_vectors_new, default_binary_entities_new = utils.gen_binary_entities_new(utils.default_nb)


default_entities_rows = utils.gen_entities_rows(utils.default_nb)
default_raw_binary_vectors_rows, default_binary_entities_rows = utils.gen_binary_entities_rows(utils.default_nb)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='common.proto',
  package='milvus.proto.common',
  syntax='proto3',
  serialized_options=_b('Z@github.com/zilliztech/milvus-distributed/internal/proto/commonpb'),
  serialized_pb=_b('\n\x0c\x63ommon.proto\x12\x13milvus.proto.common\"L\n\x06Status\x12\x32\n\nerror_code\x18\x01 \x01(\x0e\x32\x1e.milvus.proto.common.ErrorCode\x12\x0e\n\x06reason\x18\x02 \x01(\t\"*\n\x0cKeyValuePair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x15\n\x04\x42lob\x12\r\n\x05value\x18\x01 \x01(\x0c\"#\n\x07\x41\x64\x64ress\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\"m\n\x07MsgBase\x12.\n\x08msg_type\x18\x01 \x01(\x0e\x32\x1c.milvus.proto.common.MsgType\x12\r\n\x05msgID\x18\x02 \x01(\x03\x12\x11\n\ttimestamp\x18\x03 \x01(\x04\x12\x10\n\x08sourceID\x18\x04 \x01(\x03\"7\n\tMsgHeader\x12*\n\x04\x62\x61se\x18\x01 \x01(\x0b\x32\x1c.milvus.proto.common.MsgBase*\xa5\x04\n\tErrorCode\x12\x0b\n\x07Success\x10\x00\x12\x13\n\x0fUnexpectedError\x10\x01\x12\x11\n\rConnectFailed\x10\x02\x12\x14\n\x10PermissionDenied\x10\x03\x12\x17\n\x13\x43ollectionNotExists\x10\x04\x12\x13\n\x0fIllegalArgument\x10\x05\x12\x14\n\x10IllegalDimension\x10\x07\x12\x14\n\x10IllegalIndexType\x10\x08\x12\x19\n\x15IllegalCollectionName\x10\t\x12\x0f\n\x0bIllegalTOPK\x10\n\x12\x14\n\x10IllegalRowRecord\x10\x0b\x12\x13\n\x0fIllegalVectorID\x10\x0c\x12\x17\n\x13IllegalSearchResult\x10\r\x12\x10\n\x0c\x46ileNotFound\x10\x0e\x12\x0e\n\nMetaFailed\x10\x0f\x12\x0f\n\x0b\x43\x61\x63heFailed\x10\x10\x12\x16\n\x12\x43\x61nnotCreateFolder\x10\x11\x12\x14\n\x10\x43\x61nnotCreateFile\x10\x12\x12\x16\n\x12\x43\x61nnotDeleteFolder\x10\x13\x12\x14\n\x10\x43\x61nnotDeleteFile\x10\x14\x12\x13\n\x0f\x42uildIndexError\x10\x15\x12\x10\n\x0cIllegalNLIST\x10\x16\x12\x15\n\x11IllegalMetricType\x10\x17\x12\x0f\n\x0bOutOfMemory\x10\x18\x12\x11\n\rIndexNotExist\x10\x19\x12\x12\n\rDDRequestRace\x10\xe8\x07*e\n\nIndexState\x12\x12\n\x0eIndexStateNone\x10\x00\x12\x0c\n\x08Unissued\x10\x01\x12\x0e\n\nInProgress\x10\x02\x12\x0c\n\x08\x46inished\x10\x03\x12\n\n\x06\x46\x61iled\x10\x04\x12\x0b\n\x07\x44\x65leted\x10\x05*X\n\x0cSegmentState\x12\x14\n\x10SegmentStateNone\x10\x00\x12\x0c\n\x08NotExist\x10\x01\x12\x0b\n\x07Growing\x10\x02\x12\n\n\x06Sealed\x10\x03\x12\x0b\n\x07\x46lushed\x10\x04*\xfa\x05\n\x07MsgType\x12\r\n\tUndefined\x10\x00\x12\x14\n\x10\x43reateCollection\x10\x64\x12\x12\n\x0e\x44ropCollection\x10\x65\x12\x11\n\rHasCollection\x10\x66\x12\x16\n\x12\x44\x65scribeCollection\x10g\x12\x13\n\x0fShowCollections\x10h\x12\x14\n\x10GetSystemConfigs\x10i\x12\x12\n\x0eLoadCollection\x10j\x12\x15\n\x11ReleaseCollection\x10k\x12\x14\n\x0f\x43reatePartition\x10\xc8\x01\x12\x12\n\rDropPartition\x10\xc9\x01\x12\x11\n\x0cHasPartition\x10\xca\x01\x12\x16\n\x11\x44\x65scribePartition\x10\xcb\x01\x12\x13\n\x0eShowPartitions\x10\xcc\x01\x12\x13\n\x0eLoadPartitions\x10\xcd\x01\x12\x16\n\x11ReleasePartitions\x10\xce\x01\x12\x11\n\x0cShowSegments\x10\xfa\x01\x12\x14\n\x0f\x44\x65scribeSegment\x10\xfb\x01\x12\x10\n\x0b\x43reateIndex\x10\xac\x02\x12\x12\n\rDescribeIndex\x10\xad\x02\x12\x0e\n\tDropIndex\x10\xae\x02\x12\x0b\n\x06Insert\x10\x90\x03\x12\x0b\n\x06\x44\x65lete\x10\x91\x03\x12\n\n\x05\x46lush\x10\x92\x03\x12\x0b\n\x06Search\x10\xf4\x03\x12\x11\n\x0cSearchResult\x10\xf5\x03\x12\x12\n\rGetIndexState\x10\xf6\x03\x12\x1c\n\x17GetCollectionStatistics\x10\xf7\x03\x12\x1b\n\x16GetPartitionStatistics\x10\xf8\x03\x12\x10\n\x0bSegmentInfo\x10\xd8\x04\x12\r\n\x08TimeTick\x10\xb0\t\x12\x13\n\x0eQueryNodeStats\x10\xb1\t\x12\x0e\n\tLoadIndex\x10\xb2\t\x12\x0e\n\tRequestID\x10\xb3\t\x12\x0f\n\nRequestTSO\x10\xb4\t\x12\x14\n\x0f\x41llocateSegment\x10\xb5\t\x12\x16\n\x11SegmentStatistics\x10\xb6\t\x12\x15\n\x10SegmentFlushDone\x10\xb7\tBBZ@github.com/zilliztech/milvus-distributed/internal/proto/commonpbb\x06proto3')
)

_ERRORCODE = _descriptor.EnumDescriptor(
  name='ErrorCode',
  full_name='milvus.proto.common.ErrorCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Success', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UnexpectedError', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ConnectFailed', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PermissionDenied', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CollectionNotExists', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalArgument', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalDimension', index=6, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalIndexType', index=7, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalCollectionName', index=8, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalTOPK', index=9, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalRowRecord', index=10, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalVectorID', index=11, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalSearchResult', index=12, number=13,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FileNotFound', index=13, number=14,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MetaFailed', index=14, number=15,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CacheFailed', index=15, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CannotCreateFolder', index=16, number=17,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CannotCreateFile', index=17, number=18,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CannotDeleteFolder', index=18, number=19,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CannotDeleteFile', index=19, number=20,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BuildIndexError', index=20, number=21,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalNLIST', index=21, number=22,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IllegalMetricType', index=22, number=23,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OutOfMemory', index=23, number=24,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IndexNotExist', index=24, number=25,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DDRequestRace', index=25, number=1000,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=388,
  serialized_end=937,
)
_sym_db.RegisterEnumDescriptor(_ERRORCODE)

ErrorCode = enum_type_wrapper.EnumTypeWrapper(_ERRORCODE)
_INDEXSTATE = _descriptor.EnumDescriptor(
  name='IndexState',
  full_name='milvus.proto.common.IndexState',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='IndexStateNone', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Unissued', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='InProgress', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Finished', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Failed', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Deleted', index=5, number=5,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=939,
  serialized_end=1040,
)
_sym_db.RegisterEnumDescriptor(_INDEXSTATE)

IndexState = enum_type_wrapper.EnumTypeWrapper(_INDEXSTATE)
_SEGMENTSTATE = _descriptor.EnumDescriptor(
  name='SegmentState',
  full_name='milvus.proto.common.SegmentState',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SegmentStateNone', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NotExist', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Growing', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Sealed', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Flushed', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1042,
  serialized_end=1130,
)
_sym_db.RegisterEnumDescriptor(_SEGMENTSTATE)

SegmentState = enum_type_wrapper.EnumTypeWrapper(_SEGMENTSTATE)
_MSGTYPE = _descriptor.EnumDescriptor(
  name='MsgType',
  full_name='milvus.proto.common.MsgType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Undefined', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CreateCollection', index=1, number=100,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DropCollection', index=2, number=101,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HasCollection', index=3, number=102,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DescribeCollection', index=4, number=103,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ShowCollections', index=5, number=104,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GetSystemConfigs', index=6, number=105,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LoadCollection', index=7, number=106,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ReleaseCollection', index=8, number=107,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CreatePartition', index=9, number=200,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DropPartition', index=10, number=201,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HasPartition', index=11, number=202,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DescribePartition', index=12, number=203,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ShowPartitions', index=13, number=204,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LoadPartitions', index=14, number=205,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ReleasePartitions', index=15, number=206,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ShowSegments', index=16, number=250,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DescribeSegment', index=17, number=251,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CreateIndex', index=18, number=300,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DescribeIndex', index=19, number=301,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DropIndex', index=20, number=302,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Insert', index=21, number=400,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Delete', index=22, number=401,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Flush', index=23, number=402,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Search', index=24, number=500,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SearchResult', index=25, number=501,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GetIndexState', index=26, number=502,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GetCollectionStatistics', index=27, number=503,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GetPartitionStatistics', index=28, number=504,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SegmentInfo', index=29, number=600,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TimeTick', index=30, number=1200,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='QueryNodeStats', index=31, number=1201,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LoadIndex', index=32, number=1202,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RequestID', index=33, number=1203,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RequestTSO', index=34, number=1204,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AllocateSegment', index=35, number=1205,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SegmentStatistics', index=36, number=1206,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SegmentFlushDone', index=37, number=1207,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1133,
  serialized_end=1895,
)
_sym_db.RegisterEnumDescriptor(_MSGTYPE)

MsgType = enum_type_wrapper.EnumTypeWrapper(_MSGTYPE)
Success = 0
UnexpectedError = 1
ConnectFailed = 2
PermissionDenied = 3
CollectionNotExists = 4
IllegalArgument = 5
IllegalDimension = 7
IllegalIndexType = 8
IllegalCollectionName = 9
IllegalTOPK = 10
IllegalRowRecord = 11
IllegalVectorID = 12
IllegalSearchResult = 13
FileNotFound = 14
MetaFailed = 15
CacheFailed = 16
CannotCreateFolder = 17
CannotCreateFile = 18
CannotDeleteFolder = 19
CannotDeleteFile = 20
BuildIndexError = 21
IllegalNLIST = 22
IllegalMetricType = 23
OutOfMemory = 24
IndexNotExist = 25
DDRequestRace = 1000
IndexStateNone = 0
Unissued = 1
InProgress = 2
Finished = 3
Failed = 4
Deleted = 5
SegmentStateNone = 0
NotExist = 1
Growing = 2
Sealed = 3
Flushed = 4
Undefined = 0
CreateCollection = 100
DropCollection = 101
HasCollection = 102
DescribeCollection = 103
ShowCollections = 104
GetSystemConfigs = 105
LoadCollection = 106
ReleaseCollection = 107
CreatePartition = 200
DropPartition = 201
HasPartition = 202
DescribePartition = 203
ShowPartitions = 204
LoadPartitions = 205
ReleasePartitions = 206
ShowSegments = 250
DescribeSegment = 251
CreateIndex = 300
DescribeIndex = 301
DropIndex = 302
Insert = 400
Delete = 401
Flush = 402
Search = 500
SearchResult = 501
GetIndexState = 502
GetCollectionStatistics = 503
GetPartitionStatistics = 504
SegmentInfo = 600
TimeTick = 1200
QueryNodeStats = 1201
LoadIndex = 1202
RequestID = 1203
RequestTSO = 1204
AllocateSegment = 1205
SegmentStatistics = 1206
SegmentFlushDone = 1207



_STATUS = _descriptor.Descriptor(
  name='Status',
  full_name='milvus.proto.common.Status',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error_code', full_name='milvus.proto.common.Status.error_code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reason', full_name='milvus.proto.common.Status.reason', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=37,
  serialized_end=113,
)


_KEYVALUEPAIR = _descriptor.Descriptor(
  name='KeyValuePair',
  full_name='milvus.proto.common.KeyValuePair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='milvus.proto.common.KeyValuePair.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='milvus.proto.common.KeyValuePair.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=115,
  serialized_end=157,
)


_BLOB = _descriptor.Descriptor(
  name='Blob',
  full_name='milvus.proto.common.Blob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='milvus.proto.common.Blob.value', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=159,
  serialized_end=180,
)


_ADDRESS = _descriptor.Descriptor(
  name='Address',
  full_name='milvus.proto.common.Address',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ip', full_name='milvus.proto.common.Address.ip', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port', full_name='milvus.proto.common.Address.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=182,
  serialized_end=217,
)


_MSGBASE = _descriptor.Descriptor(
  name='MsgBase',
  full_name='milvus.proto.common.MsgBase',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='msg_type', full_name='milvus.proto.common.MsgBase.msg_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='msgID', full_name='milvus.proto.common.MsgBase.msgID', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='milvus.proto.common.MsgBase.timestamp', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sourceID', full_name='milvus.proto.common.MsgBase.sourceID', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=219,
  serialized_end=328,
)


_MSGHEADER = _descriptor.Descriptor(
  name='MsgHeader',
  full_name='milvus.proto.common.MsgHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='base', full_name='milvus.proto.common.MsgHeader.base', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=330,
  serialized_end=385,
)

_STATUS.fields_by_name['error_code'].enum_type = _ERRORCODE
_MSGBASE.fields_by_name['msg_type'].enum_type = _MSGTYPE
_MSGHEADER.fields_by_name['base'].message_type = _MSGBASE
DESCRIPTOR.message_types_by_name['Status'] = _STATUS
DESCRIPTOR.message_types_by_name['KeyValuePair'] = _KEYVALUEPAIR
DESCRIPTOR.message_types_by_name['Blob'] = _BLOB
DESCRIPTOR.message_types_by_name['Address'] = _ADDRESS
DESCRIPTOR.message_types_by_name['MsgBase'] = _MSGBASE
DESCRIPTOR.message_types_by_name['MsgHeader'] = _MSGHEADER
DESCRIPTOR.enum_types_by_name['ErrorCode'] = _ERRORCODE
DESCRIPTOR.enum_types_by_name['IndexState'] = _INDEXSTATE
DESCRIPTOR.enum_types_by_name['SegmentState'] = _SEGMENTSTATE
DESCRIPTOR.enum_types_by_name['MsgType'] = _MSGTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Status = _reflection.GeneratedProtocolMessageType('Status', (_message.Message,), {
  'DESCRIPTOR' : _STATUS,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:milvus.proto.common.Status)
  })
_sym_db.RegisterMessage(Status)

KeyValuePair = _reflection.GeneratedProtocolMessageType('KeyValuePair', (_message.Message,), {
  'DESCRIPTOR' : _KEYVALUEPAIR,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:milvus.proto.common.KeyValuePair)
  })
_sym_db.RegisterMessage(KeyValuePair)

Blob = _reflection.GeneratedProtocolMessageType('Blob', (_message.Message,), {
  'DESCRIPTOR' : _BLOB,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:milvus.proto.common.Blob)
  })
_sym_db.RegisterMessage(Blob)

Address = _reflection.GeneratedProtocolMessageType('Address', (_message.Message,), {
  'DESCRIPTOR' : _ADDRESS,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:milvus.proto.common.Address)
  })
_sym_db.RegisterMessage(Address)

MsgBase = _reflection.GeneratedProtocolMessageType('MsgBase', (_message.Message,), {
  'DESCRIPTOR' : _MSGBASE,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:milvus.proto.common.MsgBase)
  })
_sym_db.RegisterMessage(MsgBase)

MsgHeader = _reflection.GeneratedProtocolMessageType('MsgHeader', (_message.Message,), {
  'DESCRIPTOR' : _MSGHEADER,
  '__module__' : 'common_pb2'
  # @@protoc_insertion_point(class_scope:milvus.proto.common.MsgHeader)
  })
_sym_db.RegisterMessage(MsgHeader)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import abc
import threading

from .abstract import QueryResult
from .exceptions import BaseException
from .types import Status


# TODO: remove this to a common util
def _parameter_is_empty(func):
    import inspect
    sig = inspect.signature(func)
    # params = sig.parameters
    # todo: add more check to parameter, such as `default parameter`,
    #  `positional-only`, `positional-or-keyword`, `keyword-only`, `var-positional`, `var-keyword`
    # if len(params) == 0:
    #     return True
    # for param in params.values():
    #     if (param.kind == inspect.Parameter.POSITIONAL_ONLY or
    #         param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD) and \
    #             param.default == inspect._empty:
    #         return False
    return len(sig.parameters) == 0


class AbstractFuture:
    @abc.abstractmethod
    def result(self, **kwargs):
        '''Return deserialized result.

        It's a synchronous interface. It will wait executing until
        server respond or timeout occur(if specified).

        This API is thread-safe.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def cancel(self):
        '''Cancle gRPC future.

        This API is thread-safe.
        '''
        raise NotImplementedError()

    @abc.abstractmethod
    def done(self):
        '''Wait for request done.

        This API is thread-safe.
        '''
        raise NotImplementedError()


class Future(AbstractFuture):
    def __init__(self, future, done_callback=None, pre_exception=None):
        self._future = future
        self._done_cb = done_callback  # keep compatible (such as Future(future, done_callback)), deprecated later
        self._done_cb_list = []
        self.add_callback(done_callback)
        self._condition = threading.Condition()
        self._canceled = False
        self._done = False
        self._response = None
        self._results = None
        self._exception = pre_exception

        # self.__init()

    def add_callback(self, func):
        self._done_cb_list.append(func)

    def __del__(self):
        self._future = None

    @abc.abstractmethod
    def on_response(self, response):
        ''' Parse response from gRPC server and return results.
        '''
        raise NotImplementedError()

    def __init(self):
        ''' Register request done callback of gRPC future
        Callback function can be executed in individual subthread of gRPC, so
        there need to notify main thread when callback function finished.
        '''

        def async_done_callback(future):
            with self._condition:
                # delete gRCP future manually
                # self._future.__del__()
                # self._future = None

                # If user specify done callback function, execute it.
                try:
                    self._response = future.result()
                    self._results = self.on_response(self._response)
                    if self._done_cb:
                        if isinstance(self._results, tuple):
                            self._done_cb(*self._results)
                        elif self._results is not None:
                            self._done_cb(self._results)
                        else:
                            self._done_cb()
                except Exception as e:
                    self._exception = e
                finally:
                    self._done = True
                    self._condition.notify_all()

        if self._future:
            self._future.add_done_callback(async_done_callback)

    def result(self, **kwargs):
        self.exception()
        with self._condition:
            # future not finished. wait callback being called.
            to = kwargs.get("timeout", None)
            if self._future and not self._future.done() or not self._response:
                self._response = self._future.result(timeout=to)
                self._results = self.on_response(self._response)

                for cb in self._done_cb_list:
                    if cb:
                        # necessary to check parameter signature of cb?
                        if isinstance(self._results, tuple):
                            cb(*self._results)
                        elif _parameter_is_empty(cb):
                            cb()
                        elif self._results is not None:
                            cb(self._results)
                        else:
                            raise BaseException(1, "callback function is not legal!")

            self._condition.notify_all()

        self.exception()
        if kwargs.get("raw", False) is True:
            # just return response object received from gRPC
            return self._response

        if self._results:
            return self._results
        else:
            return self.on_response(self._response)

    # def result(self, **kwargs):
    #     self.exception()
    #     with self._condition:
    #         # future not finished. wait callback being called.
    #         to = kwargs.get("timeout", None)
    #         if not self._future.done() or not self._response:
    #             self._response = self._future.result(timeout=to)
    #         # if not self._done and not self._canceled:
    #         #     to = kwargs.get("timeout", None)
    #         #     self._condition.wait(to)
    #         #
    #         #     if not self._done and not self._canceled:
    #         #         self._condition.notify_all()
    #         # raise FutureTimeoutError("Wait timeout")

    #         self._condition.notify_all()

    #     self.exception()
    #     if kwargs.get("raw", False) is True:
    #         # just return response object received from gRPC
    #         return self._response

    #     if self._results:
    #         return self._results
    #     else:
    #         return self.on_response(self._response)

    def cancel(self):
        with self._condition:
            if self._future:
                self._future.cancel()
            # if not self._canceled or self._done:
            #     self._future.cancel()
            #     self._canceled = True
            self._condition.notify_all()

    def is_done(self):
        return self._done

    def done(self):
        # self.exception()
        with self._condition:
            if self._future and not self._future.done():
                try:
                    self._future.result()
                except Exception as e:
                    self._exception = e

            self._condition.notify_all()

    def exception(self):
        if self._exception:
            raise self._exception
        if self._future and not self._future.done():
            self._future.exception()


class SearchFuture(Future):
    def __init__(self, future, done_callback=None, auto_id=True, pre_exception=None):
        super().__init__(future, done_callback, pre_exception)
        self._auto_id = auto_id

    def on_response(self, response):
        if response.status.error_code == 0:
            return QueryResult(response, self._auto_id)

        status = response.status
        raise BaseException(status.error_code, status.reason)


class InsertFuture(Future):
    def on_response(self, response):
        status = response.status
        if status.error_code == 0:
            return list(range(response.rowID_begin, response.rowID_end))
            # return list(response.entity_id_array)

        status = response.status
        raise BaseException(status.error_code, status.reason)


class CreateIndexFuture(Future):
    def on_response(self, response):
        if response.error_code != 0:
            raise BaseException(response.error_code, response.reason)

        return Status(response.error_code, response.reason)


class CreateFlatIndexFuture(AbstractFuture):
    def __init__(self, res, done_callback=None, pre_exception=None):
        self._results = res
        self._done_cb = done_callback
        self._done_cb_list = []
        self.add_callback(done_callback)
        self._condition = threading.Condition()
        self._exception = pre_exception

    def add_callback(self, func):
        self._done_cb_list.append(func)

    def __del__(self):
        self._results = None

    def on_response(self, response):
        pass

    def result(self, **kwargs):
        self.exception()
        with self._condition:
            for cb in self._done_cb_list:
                if cb:
                    # necessary to check parameter signature of cb?
                    if isinstance(self._results, tuple):
                        cb(*self._results)
                    elif _parameter_is_empty(cb):
                        cb()
                    elif self._results is not None:
                        cb(self._results)
                    else:
                        raise BaseException(1, "callback function is not legal!")
            return self._results

    def cancel(self):
        with self._condition:
            self._condition.notify_all()

    def is_done(self):
        return True

    def done(self):
        with self._condition:
            self._condition.notify_all()

    def exception(self):
        if self._exception:
            raise self._exception


class FlushFuture(Future):
    def on_response(self, response):
        if response.error_code != 0:
            raise BaseException(response.error_code, response.reason)


class LoadCollectionFuture(Future):
    def on_response(self, response):
        if response.error_code != 0:
            raise BaseException(response.error_code, response.reason)


class LoadPartitionsFuture(Future):
    def on_response(self, response):
        if response.error_code != 0:
            raise BaseException(response.error_code, response.reason)



import sys
import datetime
from urllib.parse import urlparse
from .exceptions import ParamError


def is_legal_host(host):
    if not isinstance(host, str):
        return False

    return True


def is_legal_port(port):
    if isinstance(port, (str, int)):
        try:
            _port = int(port)
        except ValueError:
            return False
        else:
            if _port <= 0 or _port > 65535:
                return False

    return True


def is_legal_uri(uri):
    if uri is None:
        return True

    try:
        _uri = urlparse(uri)
        return _uri.scheme == 'tcp'
    except (AttributeError, ValueError, TypeError):
        return False


def is_legal_vector(array):
    if not array or \
            not isinstance(array, list) or \
            len(array) == 0:
        return False

    # for v in array:
    #     if not isinstance(v, float):
    #         return False

    return True


def is_legal_bin_vector(array):
    if not array or \
            not isinstance(array, bytes) or \
            len(array) == 0:
        return False

    return True


def is_legal_numpy_array(array):
    return False if array is None or array.size == 0 else True


# def is_legal_records(value):
#     param_error = ParamError('A vector must be a non-empty, 2-dimensional array and '
#                              'must contain only elements with the float data type or the bytes data type.')
#
#     if isinstance(value, np.ndarray):
#         if not is_legal_numpy_array(value):
#             raise param_error
#
#         return True
#
#     if not isinstance(value, list) or len(value) == 0:
#         raise param_error
#
#     if isinstance(value[0], bytes):
#         check_func = is_legal_bin_vector
#     elif isinstance(value[0], list):
#         check_func = is_legal_vector
#     else:
#         raise param_error
#
#     _dim = len(value[0])
#     for record in value:
#         if not check_func(record):
#             raise param_error
#         if _dim != len(record):
#             raise ParamError('Whole vectors must have the same dimension')
#
#     return True


def int_or_str(item):
    if isinstance(item, int):
        return str(item)

    return item


def is_correct_date_str(param):
    try:
        datetime.datetime.strptime(param, '%Y-%m-%d')
    except ValueError:
        raise ParamError('Incorrect data format, should be YYYY-MM-DD')

    return True


def is_legal_dimension(dim):
    return isinstance(dim, int)


def is_legal_index_size(index_size):
    return isinstance(index_size, int)


def is_legal_table_name(table_name):
    return isinstance(table_name, str) and len(table_name) > 0


def is_legal_field_name(field_name):
    return isinstance(field_name, str) and len(field_name) > 0


def is_legal_nlist(nlist):
    return not isinstance(nlist, bool) and isinstance(nlist, int)


def is_legal_topk(topk):
    return not isinstance(topk, bool) and isinstance(topk, int)


def is_legal_ids(ids):
    if not isinstance(ids, list) or \
            len(ids) == 0:
        return False

    # TODO: Here check id valid value range may not match other SDK
    for i in ids:
        if not isinstance(i, (int, str)):
            return False
        try:
            i_ = int(i)
            if i_ < 0 or i_ > sys.maxsize:
                return False
        except:
            return False

    return True


def is_legal_nprobe(nprobe):
    return isinstance(nprobe, int)


def is_legal_cmd(cmd):
    return isinstance(cmd, str) and len(cmd) > 0


def parser_range_date(date):
    if isinstance(date, datetime.date):
        return date.strftime('%Y-%m-%d')

    if isinstance(date, str):
        if not is_correct_date_str(date):
            raise ParamError('Date string should be YY-MM-DD format!')

        return date

    raise ParamError(
        'Date should be YY-MM-DD format string or datetime.date, '
        'or datetime.datetime object')


def is_legal_date_range(start, end):
    start_date = datetime.datetime.strptime(start, "%Y-%m-%d")
    end_date = datetime.datetime.strptime(end, "%Y-%m-%d")
    if (end_date - start_date).days < 0:
        return False

    return True


def is_legal_partition_tag(tag):
    return tag is not None and isinstance(tag, str)


def is_legal_partition_tag_array(tag_array):
    if tag_array is None:
        return True

    if not isinstance(tag_array, list):
        return False

    for tag in tag_array:
        if not is_legal_partition_tag(tag):
            return False

    return True


# https://milvus.io/cn/docs/v1.0.0/metric.md#floating
def is_legal_index_metric_type(index_type: str, metric_type: str) -> bool:
    if index_type not in ["FLAT",
                          "IVF_FLAT",
                          "IVF_SQ8",
                          # "IVF_SQ8_HYBRID",
                          "IVF_PQ",
                          "HNSW",
                          # "NSG",
                          "ANNOY",
                          "RHNSW_FLAT",
                          "RHNSW_PQ",
                          "RHNSW_SQ", ]:
        return False
    if metric_type not in ["L2", "IP"]:
        return False
    return True


# https://milvus.io/cn/docs/v1.0.0/metric.md#binary
def is_legal_binary_index_metric_type(index_type: str, metric_type: str) -> bool:
    if index_type == "BIN_FLAT":
        if metric_type in ["JACCARD", "TANIMOTO", "HAMMING", "SUBSTRUCTURE", "SUPERSTRUCTURE"]:
            return True
    elif index_type == "BIN_IVF_FLAT":
        if metric_type in ["JACCARD", "TANIMOTO", "HAMMING"]:
            return True
    return False


def _raise_param_error(param_name, param_value):
    raise ParamError("`{}` value {} is illegal".format(param_name, param_value))


def check_pass_param(*args, **kwargs):
    if kwargs is None:
        raise ParamError("Param should not be None")

    for key, value in kwargs.items():
        if key in ("collection_name",):
            if not is_legal_table_name(value):
                _raise_param_error(key, value)
        elif key == "field_name":
            if not is_legal_field_name(value):
                _raise_param_error(key, value)
        elif key == "dimension":
            if not is_legal_dimension(value):
                _raise_param_error(key, value)
        elif key == "index_file_size":
            if not is_legal_index_size(value):
                _raise_param_error(key, value)
        elif key in ("topk", "top_k"):
            if not is_legal_topk(value):
                _raise_param_error(key, value)
        elif key in ("ids",):
            if not is_legal_ids(value):
                _raise_param_error(key, value)
        elif key in ("nprobe",):
            if not is_legal_nprobe(value):
                _raise_param_error(key, value)
        elif key in ("nlist",):
            if not is_legal_nlist(value):
                _raise_param_error(key, value)
        elif key in ("cmd",):
            if not is_legal_cmd(value):
                _raise_param_error(key, value)
        elif key in ("partition_tag",):
            if not is_legal_partition_tag(value):
                _raise_param_error(key, value)
        elif key in ("partition_tag_array",):
            if not is_legal_partition_tag_array(value):
                _raise_param_error(key, value)
        # elif key in ("records",):
        #     if not is_legal_records(value):
        #         _raise_param_error(key, value)
        else:
            raise ParamError("unknown param `{}`".format(key))
